from typing import Any, Callable, List, Optional, TYPE_CHECKING, Union

from .._streamingprices import StreamingPrices
from ..._tools import (
    PRICING_DATETIME_PATTERN,
    cached_property,
    create_repr,
    make_callback,
)
from ..._tools._dataframe import convert_df_columns_to_datetime_use_re_compile
from ...delivery._stream.base_stream import StreamOpenWithUpdatesMixin

if TYPE_CHECKING:
    import pandas
    from ... import OpenState
    from ...content._streamingprices import PricingStream
    from ..._core.session import Session


class Stream(StreamOpenWithUpdatesMixin):
    """
    Summary line of this class are used for requesting, processing and managing a set of
    streaming level 1 (MarketPrice domain) quotes and trades

    Extended description of this class:
        The object automatically manages a set of streaming caches available for access
        at any time. Your application can then reach into this cache and pull out
        real-time fields by just calling a simple access method.
        The object also emits a number of different events, your application can
        listen to in order to be notified of the latest field values in real-times.
        The object iterable.

    Parameters
    ----------
    session : Session, optional
        Means default session would be used
    universe : str or list of str, optional
        The single/multiple instrument/s name (e.g. "EUR=" or ["EUR=", "CAD=", "UAH="])
    fields : list, optional
        Specifies the specific fields to be delivered when messages arrive
    service : str, optional
        Name of the streaming service publishing the instruments
    extended_params : dict, optional
        If necessary other parameters

    Examples
    --------
    >>> from refinitiv.data.content import pricing
    >>> definition = pricing.Definition("EUR=")
    >>> stream = definition.get_stream()
    """

    def __init__(
        self,
        session: "Session" = None,
        universe: Union[str, List[str]] = None,
        fields: Optional[list] = None,
        service: Optional[str] = None,
        extended_params: Optional[dict] = None,
    ) -> None:
        self._session = session
        self._universe = universe or []
        self._fields = fields
        self._service = service
        self._extended_params = extended_params

    @cached_property
    def _stream(self) -> StreamingPrices:
        return StreamingPrices(
            universe=self._universe,
            session=self._session,
            fields=self._fields,
            service=self._service,
            extended_params=self._extended_params,
        )

    def open(self, with_updates: bool = True) -> "OpenState":
        """
        Summary line of this func open the Pricing connection by sending corresponding
        requests for all requested instruments

        Extended description of this class:
             It will be opened once all the requested instruments are received
             either on_refresh, on_status or other method.
             Then the pricing.Stream can be used in order to retrieve data.

        Parameters
        ----------
        with_updates : bool, optional
            actions:
                True - the streaming will work as usual and the data will be received
                continuously.
                False - only one data snapshot will be received
                (single Refresh 'NonStreaming') and stream will be closed automatically.

            Default: True

        Returns
        -------
        OpenState

        Examples
        --------
        >>> from refinitiv.data.content import pricing
        >>> definition = pricing.Definition("EUR=")
        >>> stream = definition.get_stream()
        >>> stream.open()
        """
        self._stream.open(with_updates=with_updates)
        return self.open_state

    def close(self) -> "OpenState":
        """
        Summary line of this func closes the pricing connection, releases resources

        Returns
        -------
        OpenState

        Examples
        --------
        >>> from refinitiv.data.content import pricing
        >>> definition = pricing.Definition("EUR=")
        >>> stream = definition.get_stream()
        >>> stream.open()
        >>> stream.close()
        """
        self._stream.close()
        return self.open_state

    def _get_fields(self, universe: str, fields: Optional[list] = None) -> dict:
        """
        Returns a dict of the fields for a requested universe.

        Parameters
        ----------
        universe: str
            Name of the instrument (e.g. 'EUR=' ...).
        fields: list, optional
            The fields that are listed in the `fields` parameter.

        Returns
        -------
        dict

        Examples
        --------
        >>> from refinitiv.data.content import pricing
        >>> definition = pricing.Definition(["AIRP.PA", "AIR.PA", "ALSO.PA"])
        >>> stream = definition.get_stream()
        >>> fields
        ... {'AIRP.PA': {'BID': 147.14, 'ASK': 147.16}}
        """
        _fields = {
            universe: {
                key: value
                for key, value in self._stream[universe].items()
                if fields is None or key in fields
            }
        }
        return _fields

    def get_snapshot(
        self,
        universe: Union[str, List[str], None] = None,
        fields: Optional[List[str]] = None,
        convert: bool = True,
    ) -> "pandas.DataFrame":
        """
        Returns a Dataframe filled with snapshot values for a list of instrument names
        and a list of fields.

        Parameters
        ----------
        universe: str, list of str, optional
            List of instruments to request snapshot data on.

        fields: list of str, optional
            List of fields to request.

        convert: bool, optional
            If True, force numeric conversion for all values.
            By default True.

        Returns
        -------
            pandas.DataFrame

            pandas.DataFrame content:
                - columns : instrument and fieled names
                - rows : instrument name and field values

        Raises
        ------
            Exception
                If request fails or if server returns an error

            ValueError
                If a parameter type or value is wrong

        Examples
        --------
        >>> from refinitiv.data.content import pricing
        >>> definition = pricing.Definition(
        ...    ["MSFT.O", "GOOG.O", "IBM.N"],
        ...    fields=["BID", "ASK", "OPEN_PRC"]
        ...)
        >>> stream = definition.get_stream()
        >>> data = stream.get_snapshot(["MSFT.O", "GOOG.O"], ["BID", "ASK"])
        >>> data
        ... "      Instrument   BID       ASK        "
        ... "0     MSFT.O        150.9000   150.9500 "
        ... "1     GOOG.O        1323.9000  1327.7900"
        """
        df = self._stream.get_snapshot(
            universe=universe, fields=fields, convert=convert
        )
        convert_df_columns_to_datetime_use_re_compile(df, PRICING_DATETIME_PATTERN)
        return df

    def on_refresh(self, func: Callable[[Any, str, "Stream"], Any]) -> "Stream":
        """
        Called when a stream on instrument_name was opened successfully or when the
        stream is refreshed by the server
        This callback is called with the reference to the steam object, the instrument
        name and the instrument full image

        Parameters
        ----------
        func : Callable
            Callable object to process retrieved data

        Returns
        -------
        current instance

        Examples
        --------
        >>> import datetime
        >>> from refinitiv.data.content import pricing
        >>> definition = pricing.Definition("EUR=")
        >>> stream = definition.get_stream()
        >>> def on_refresh(message, ric, stream):
        ...    current_time = datetime.datetime.now().time()
        ...    print("\t{} | Receive refresh [{}] : {}".format(current_time, ric, message))  # noqa
        >>> stream.on_refresh(on_refresh)
        >>> stream.open()
        """
        self._stream.on_refresh(make_callback(func))
        return self

    def on_update(self, func: Callable[[Any, str, "Stream"], Any]) -> "Stream":
        """
        Called when an update is received for a instrument_name
        This callback is called with the reference to the stream object, the instrument
        name and the instrument update

        Parameters
        ----------
        func : Callable
            Callable object to process retrieved data

        Returns
        -------
        current instance

        Examples
        --------
        >>> import datetime
        >>> from refinitiv.data.content import pricing
        >>> definition = pricing.Definition("EUR=")
        >>> stream = definition.get_stream()
        >>> def on_update(update, ric, stream):
        ...     current_time = datetime.datetime.now().time()
        ...     print("\t{} | Receive update [{}] : {}".format(current_time, ric, update))  # noqa
        >>> stream.on_update(on_update)
        >>> stream.open()
        """
        self._stream.on_update(make_callback(func))
        return self

    def on_status(self, func: Callable[[Any, str, "Stream"], Any]) -> "Stream":
        """
        Called when a status is received for a instrument_name
        This callback is called with the reference to the stream object, the instrument
        name and the instrument status

        Parameters
        ----------
        func : Callable
            Callable object to process retrieved data

        Returns
        -------
        current instance

        Examples
        --------
        >>> from refinitiv.data.content import pricing
        >>> definition = pricing.Definition("EUR=")
        >>> stream = definition.get_stream()
        >>> def on_status(status, ric, stream):
        ...     print("\tReceive status [{}] : {}".format(ric, status))
        >>> stream.on_status(on_status)
        >>> stream.open()
        """
        self._stream.on_status(make_callback(func))
        return self

    def on_complete(self, func: Callable[["Stream"], Any]) -> "Stream":
        """
        Called when all subscriptions are completed
        This callback is called with the reference to the stream object

        Parameters
        ----------
        func : Callable
            Callable object to process retrieved data

        Returns
        -------
        current instance

        Examples
        --------
        >>> import datetime
        >>> from refinitiv.data.content import pricing
        >>> definition = pricing.Definition("EUR=")
        >>> stream = definition.get_stream()
        >>> def on_complete(stream):
        ...    current_time = datetime.datetime.now().time()
        ...    print("\t{} | Receive complete".format(current_time))
        >>> stream.on_complete(on_complete)
        >>> stream.open()
        """
        self._stream.on_complete(func)
        return self

    def on_error(self, func: Callable) -> "Stream":
        self._stream.on_error(func)
        return self

    def __iter__(self):
        return self._stream.__iter__()

    def __getitem__(self, item) -> "PricingStream":
        return self._stream.__getitem__(item)

    def __len__(self) -> int:
        return self._stream.__len__()

    def __repr__(self):
        return create_repr(
            self,
            class_name=self.__class__.__name__,
            content=f"{{name='{self._universe}'}}",
        )



import asyncio
from abc import abstractmethod
from typing import Tuple, List

from ._intervals import DayIntervalType
from ._join_responses import get_first_success_response
from .._errors import RDError
from ..delivery._data import _data_provider
from ..delivery._data._data_provider import DataProvider, Response


# ---------------------------------------------------------------------------
#   Raw data parser
# ---------------------------------------------------------------------------


class ErrorParser(_data_provider.Parser):
    def process_failed_response(self, raw_response):
        parsed_data = super().process_failed_response(raw_response)
        status = parsed_data.get("status", {})
        error = status.get("error", {})
        errors = error.get("errors", [])
        error_code = parsed_data.get("error_code")
        error_message = parsed_data.get("error_message", "")
        err_msgs = []
        err_codes = []
        for err in errors:
            reason = err.get("reason")
            if reason:
                err_codes.append(error_code)
                err_msgs.append(f"{error_message}: {reason}")

        if err_msgs and err_codes:
            parsed_data["error_code"] = err_codes
            parsed_data["error_message"] = err_msgs

        return parsed_data


# ---------------------------------------------------------------------------
#   Content data validator
# ---------------------------------------------------------------------------


def get_invalid_universes(universes):
    result = []
    for universe in universes:
        if universe.get("Organization PermID") == "Failed to resolve identifier(s).":
            result.append(universe.get("Instrument"))
    return result


def get_universe_from_raw_response(raw_response):
    universe = raw_response.url.params["universe"]
    universe = universe.split(",")
    return universe


class UniverseContentValidator(_data_provider.ContentValidator):
    def validate(self, data, *args, **kwargs):
        is_valid = super().validate(data)
        if not is_valid:
            return is_valid

        content_data = data.get("content_data", {})
        error = content_data.get("error", {})
        universes = content_data.get("universe", [])
        invalid_universes = get_invalid_universes(universes)

        if error:
            is_valid = False
            data["error_code"] = error.get("code")

            error_message = error.get("description")
            if error_message == "Unable to resolve all requested identifiers.":
                universe = get_universe_from_raw_response(data["raw_response"])
                error_message = f"{error_message} Requested items: {universe}"

            if not error_message:
                error_message = error.get("message")
                errors = error.get("errors")
                if isinstance(errors, list):
                    errors = "\n".join(map(str, errors))
                    error_message = f"{error_message}:\n{errors}"
            data["error_message"] = error_message

        elif invalid_universes:
            data["error_message"] = f"Failed to resolve identifiers {invalid_universes}"

        return is_valid


# ---------------------------------------------------------------------------
#   Provider layer
# ---------------------------------------------------------------------------


class ContentProviderLayer(_data_provider.DataProviderLayer):
    def __init__(self, content_type, **kwargs):
        _data_provider.DataProviderLayer.__init__(
            self,
            data_type=content_type,
            **kwargs,
        )


# ---------------------------------------------------------------------------
#   DataProvider
# ---------------------------------------------------------------------------


class HistoricalDataProvider(DataProvider):
    @staticmethod
    @abstractmethod
    def _join_responses(
        responses: List[Tuple[str, Response]],
        fields: List[str],
        kwargs,
    ):
        pass

    async def _create_task(self, name, *args, **kwargs) -> Tuple[str, Response]:
        kwargs["universe"] = name
        response = await super().get_data_async(*args, **kwargs)
        return name, response

    def get_data(self, *args, **kwargs) -> Response:
        universe = kwargs.get("universe", [])
        fields = copy_fields(kwargs.get("fields"))
        responses = []
        for inst_name in universe:
            kwargs["universe"] = inst_name
            response = super().get_data(*args, **kwargs)
            responses.append((inst_name, response))

        validate_responses(responses)
        return self._join_responses(responses, fields, kwargs)

    async def get_data_async(self, *args, **kwargs) -> Response:
        universe = kwargs.get("universe", [])
        fields = copy_fields(kwargs.get("fields"))
        tasks = []
        for inst_name in universe:
            tasks.append(self._create_task(inst_name, *args, **kwargs))

        responses = await asyncio.gather(*tasks)
        return self._join_responses(responses, fields, kwargs)


def copy_fields(fields: List[str]) -> List[str]:
    if fields is None:
        return []

    if not isinstance(fields, list):
        raise AttributeError(f"fields not support type {type(fields)}")

    return fields[:]


def validate_responses(responses: List[Tuple[str, Response]]):
    response = get_first_success_response(responses)
    if response is None:
        error_message = "ERROR: No successful response.\n"
        for inst_name, response in responses:
            if response.errors:
                error = response.errors[0]
                error_message += "({}, {}), ".format(error.code, error.message)
        error_message = error_message[:-2]
        raise RDError(1, f"No data to return, please check errors: {error_message}")


field_timestamp_by_day_interval_type = {
    DayIntervalType.INTER: "DATE",
    DayIntervalType.INTRA: "DATE_TIME",
}
axis_by_day_interval_type = {
    DayIntervalType.INTRA: "Timestamp",
    DayIntervalType.INTER: "Date",
}



# coding: utf8

__all__ = ["ExerciseStyle"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class ExerciseStyle(Enum):
    """
    Possible values:
        - AMER
        - BERM
        - EURO
    """

    AMER = "AMER"
    BERM = "BERM"
    EURO = "EURO"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(ExerciseStyle, _EXERCISE_STYLE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_EXERCISE_STYLE_VALUES_IN_LOWER_BY_EXERCISE_STYLE, some)


_EXERCISE_STYLE_VALUES = tuple(t.value for t in ExerciseStyle)
_EXERCISE_STYLE_VALUES_IN_LOWER_BY_EXERCISE_STYLE = {
    name.lower(): item for name, item in ExerciseStyle.__members__.items()
}



# coding: utf-8


import collections
from typing import List, Union, Dict, Tuple

from ..._types import OptInt, OptStr

REF_COUNT_FIELD_NAME = "REF_COUNT"
RECORD_TYPE_FIELD_NAME = "RECORDTYPE"
RDN_DISPLAY_FIELD_NAME = "RDNDISPLAY"
PREF_DISPLAY_FIELD_NAME = "PREF_DISP"
PREV_DISPLAY_FIELD_NAME = "PREV_DISP"
PREF_LINK_FIELD_NAME = "PREF_LINK"
DISPLAY_NAME_FIELD_NAME = "DSPLY_NAME"
INVALID_DISPLAY_TEMPLATE_VALUES = ["@@@", 0]

NCharsConfig = collections.namedtuple(
    "NCharsConfig",
    [
        "name",
        "link_field_name_template",
        "start_no_link_field_name",
        "end_no_link_field_name",
        "prev_field_name",
        "next_field_name",
    ],
)

config_10_chars = NCharsConfig(
    name="10_chars",
    link_field_name_template="LINK_{:d}",
    start_no_link_field_name=1,
    end_no_link_field_name=14,
    prev_field_name="PREV_LR",
    next_field_name="NEXT_LR",
)

config_17_chars = NCharsConfig(
    name="17_chars",
    link_field_name_template="LONGLINK{:d}",
    start_no_link_field_name=1,
    end_no_link_field_name=14,
    prev_field_name="LONGPREVLR",
    next_field_name="LONGNEXTLR",
)

config_32_chars = NCharsConfig(
    name="32_chars",
    link_field_name_template="BR_LINK{:d}",
    start_no_link_field_name=1,
    end_no_link_field_name=14,
    prev_field_name="BR_PREVLR",
    next_field_name="BR_NEXTLR",
)


class InvalidChainRecordException(Exception):
    pass


class ChainRecord(object):
    def __init__(self, config: NCharsConfig, fields: dict) -> None:
        self._config: NCharsConfig = config
        self._fields = fields
        self.num_constituents: OptInt = None
        self.link_field_by_index: Dict[int, str] = {}
        self.prev_chain_record_name: OptStr = None
        self.next_chain_record_name: OptStr = None
        self.record_type: OptInt = None
        self.rdn_display: OptInt = None
        self.pref_display: OptInt = None
        self.prev_display: OptInt = None
        self.pref_link: OptStr = None
        self.display_name: OptStr = None

    def __repr__(self) -> str:
        return (
            f"ChainRecord("
            f"num_constituents={self.num_constituents}, "
            f"link_field_by_index={self.link_field_by_index}, "
            f"prev_chain_record_name={self.prev_chain_record_name}, "
            f"next_chain_record_name={self.next_chain_record_name}, "
            f"record_type={self.record_type}, "
            f"rdn_display={self.rdn_display}, "
            f"pref_display={self.pref_display}, "
            f"prev_display={self.prev_display}, "
            f"pref_link={self.pref_link}, "
            f"display_name={self.display_name})"
        )

    @property
    def constituents(self) -> List[str]:
        return [
            constituent for _, constituent in sorted(self.link_field_by_index.items())
        ]

    @property
    def display_template(self) -> int:
        if is_valid_disp_tmpl(self.pref_display):
            return self.pref_display
        elif is_valid_disp_tmpl(self.prev_display):
            return self.prev_display
        else:
            return self.rdn_display

    def update(self, fields) -> Dict[int, Tuple[str, str]]:

        get = fields.get

        old_num_constituents = None
        new_num_constituents = None

        if REF_COUNT_FIELD_NAME in fields:
            old_num_constituents = self.num_constituents
            self.num_constituents = get(REF_COUNT_FIELD_NAME)
            new_num_constituents = self.num_constituents

        self.record_type = get(RECORD_TYPE_FIELD_NAME, self.record_type)
        self.rdn_display = get(RDN_DISPLAY_FIELD_NAME, self.rdn_display)
        self.pref_display = get(PREF_DISPLAY_FIELD_NAME, self.pref_display)
        self.prev_display = get(PREV_DISPLAY_FIELD_NAME, self.prev_display)
        self.pref_link = get(PREF_LINK_FIELD_NAME, self.pref_link)
        self.display_name = get(DISPLAY_NAME_FIELD_NAME, self.display_name)

        if old_num_constituents and new_num_constituents:
            num_constituents = new_num_constituents
            if old_num_constituents > new_num_constituents:
                num_constituents = old_num_constituents

        else:
            num_constituents = self.num_constituents

        cfg = self._config

        index_to_old_and_new_constituent: Dict[int, Tuple[str, str]] = {}
        start = cfg.start_no_link_field_name
        stop = num_constituents + 1
        for i in range(start, stop):
            link_field_name = cfg.link_field_name_template.format(i)

            if link_field_name in fields:
                old_constituent = self.link_field_by_index.get(i, None)
                self.link_field_by_index[i] = get(link_field_name)
                new_constituent = self.link_field_by_index[i]
                i = i - cfg.start_no_link_field_name
                index_to_old_and_new_constituent[i] = (old_constituent, new_constituent)

        self.prev_chain_record_name = get(
            cfg.prev_field_name, self.prev_chain_record_name
        )
        self.next_chain_record_name = get(
            cfg.next_field_name, self.next_chain_record_name
        )

        return index_to_old_and_new_constituent


def _can_create(fields: dict, config: NCharsConfig) -> bool:
    if REF_COUNT_FIELD_NAME not in fields:
        return False

    start = config.start_no_link_field_name
    stop = config.end_no_link_field_name + 1
    for i in range(start, stop):
        if config.link_field_name_template.format(i) not in fields:
            return False

    prev_not_in_fields = config.prev_field_name not in fields
    next_not_in_fields = config.next_field_name not in fields
    if prev_not_in_fields and next_not_in_fields:
        return False

    return True


def can_create_chain_record(fields: dict) -> bool:
    if _can_create(fields, config_17_chars):
        is_valid = True
    elif _can_create(fields, config_10_chars):
        is_valid = True
    elif _can_create(fields, config_32_chars):
        is_valid = True
    else:
        is_valid = False

    return is_valid


def _create_chain_record(fields: dict, config: NCharsConfig) -> ChainRecord:
    chain_record: ChainRecord = ChainRecord(config, fields)
    get = fields.get
    chain_record.record_type = get(RECORD_TYPE_FIELD_NAME)
    chain_record.rdn_display = get(RDN_DISPLAY_FIELD_NAME)
    chain_record.pref_display = get(PREF_DISPLAY_FIELD_NAME)
    chain_record.prev_display = get(PREV_DISPLAY_FIELD_NAME)
    chain_record.pref_link = get(PREF_LINK_FIELD_NAME)
    chain_record.display_name = get(DISPLAY_NAME_FIELD_NAME)

    try:
        chain_record.num_constituents = fields[REF_COUNT_FIELD_NAME]
        if chain_record.num_constituents:
            start = config.start_no_link_field_name
            stop = chain_record.num_constituents + 1
            for i in range(start, stop):
                link_field_name = config.link_field_name_template.format(i)
                chain_record.link_field_by_index[i] = fields[link_field_name]

    except KeyError:
        raise InvalidChainRecordException(
            f"ERROR!!! Invalid chain record template of {config}"
        )

    chain_record.prev_chain_record_name = get(config.prev_field_name)
    chain_record.next_chain_record_name = get(config.next_field_name)
    return chain_record


def create_chain_record(fields: dict) -> ChainRecord:
    if _can_create(fields, config_17_chars):
        chain_record = _create_chain_record(fields, config_17_chars)
    elif _can_create(fields, config_10_chars):
        chain_record = _create_chain_record(fields, config_10_chars)
    elif _can_create(fields, config_32_chars):
        chain_record = _create_chain_record(fields, config_32_chars)
    else:
        raise ValueError(f"Cannot create chain record from fields={fields}")

    return chain_record


def is_valid_disp_tmpl(disp_tmpl: Union[str, int]) -> bool:
    if isinstance(disp_tmpl, int):
        is_valid = disp_tmpl not in INVALID_DISPLAY_TEMPLATE_VALUES

    elif isinstance(disp_tmpl, str):
        is_valid = disp_tmpl not in INVALID_DISPLAY_TEMPLATE_VALUES
        is_valid = is_valid and disp_tmpl.strip() != ""

    else:
        is_valid = bool(disp_tmpl)

    return is_valid



# coding: utf8

from enum import Enum, unique
from ...._enums._common_tools import (
    _convert_to_str,
    _normalize,
)


@unique
class FxBinaryType(Enum):
    DIGITAL = "Digital"
    NO_TOUCH = "NoTouch"
    NONE = "None"
    ONE_TOUCH_DEFERRED = "OneTouchDeferred"
    ONE_TOUCH_IMMEDIATE = "OneTouchImmediate"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(FxBinaryType, _FX_BINARY_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_FX_BINARY_TYPE_VALUES_IN_LOWER_BY_FX_BINARY_TYPE, some)


_FX_BINARY_TYPE_VALUES = tuple(t.value for t in FxBinaryType)
_FX_BINARY_TYPE_VALUES_IN_LOWER_BY_FX_BINARY_TYPE = {
    name.lower(): item for name, item in FxBinaryType.__members__.items()
}



from .._data._data_provider import Data
from ._tools import _get_query_parameter


class BaseData(Data):
    def __init__(self, raw, iter_object=None, **kwargs):
        super().__init__(raw, **kwargs)
        url = self.raw.get("@nextLink", None)
        self.raw["skip_token"] = _get_query_parameter("skipToken", url)
        self._iter_object = iter_object


class BucketData(BaseData):
    @property
    def buckets(self):
        return self._iter_object


class FileSetData(BaseData):
    @property
    def file_sets(self):
        return self._iter_object


class PackageData(BaseData):
    @property
    def packages(self):
        return self._iter_object


class FileData(BaseData):
    @property
    def files(self):
        return self._iter_object



import re
import sys
from collections import Counter
from typing import Callable, Optional, Union, TYPE_CHECKING

import numpy as np
import pandas as pd
from pandas import DataFrame, Series, merge

from ._tools import get_name_columns_without_datetime
from .._core.session import get_default
from .._tools import (
    ADC_FUNC_PATTERN_IN_FIELDS,
    ADC_PARAM_IN_FIELDS,
    ADC_TR_PATTERN,
    DEBUG,
    fields_arg_parser,
    universe_arg_parser,
)
from .._tools._dataframe import convert_df_columns_to_datetime
from ..content import fundamental_and_reference
from ..content.pricing._stream_facade import Stream
from ..errors import RDError

NOT_ALLOWED_SNAPSHOT_FIELDS = {"Date"}


if TYPE_CHECKING:
    from logging import Logger


def get_data(
    universe: Union[str, list],
    fields: Union[str, list, None] = None,
    parameters: Union[str, dict, None] = None,
    use_field_names_in_headers: bool = False,
) -> DataFrame:
    """
    With this tool you can request data from ADC, realtime pricing data or
    combination of both;

    Parameters
    ----------
        universe: str | list ,
            instruments to request.
        fields: str | list .
            fields to request.
        parameters: str | dict,
            Single global parameter key=value or dictionary
            of global parameters to request.
        use_field_names_in_headers: bool
            Return field name as column headers for data instead of title


    Returns
    -------
    pandas.DataFrame

     Examples
    --------
    >>> get_data(universe=['IBM.N', 'VOD.L'], fields=['BID', 'ASK'])
    >>> get_data(
    ...     universe=['GOOG.O', 'AAPL.O'],
    ...     fields=['TR.EV','TR.EVToSales'],
    ...     parameters = {'SDate': '0CY', 'Curn': 'CAD'}
    ...)
    >>> get_data(
    ...     universe=['IBM.N', 'VOD.L'],
    ...     fields=['BID', 'ASK', 'TR.Revenue'],
    ...     parameters = {'SCale':6, 'SDate':0, 'EDate':-3, 'FRQ':'FY'}
    ...)
    """
    if isinstance(universe, str):
        universe = [universe]
    return _get_data(
        _fundamental_data=fundamental_and_reference.Definition,
        _pricing_stream=Stream(universe=universe, fields=fields),
        universe=universe,
        fields=fields,
        parameters=parameters,
        use_field_names_in_headers=use_field_names_in_headers,
    )


def _get_data(
    _fundamental_data: Callable, _pricing_stream: Stream, **kwargs
) -> DataFrame:
    universe = kwargs.pop("universe")
    logger = get_default().logger()
    fields = kwargs.pop("fields")
    use_field_names_in_headers = kwargs.pop("use_field_names_in_headers", False)

    adc_df = DataFrame()
    _add_flag(adc_df)

    pricing_df = DataFrame()
    _add_flag(pricing_df)

    fields = fields_arg_parser.get_list(fields) if fields else []
    parameters = kwargs.pop("parameters")

    adc_tr_fields = [i for i in fields if re.match(ADC_TR_PATTERN, i)]
    adc_funcs_in_fields = [i for i in fields if re.match(ADC_FUNC_PATTERN_IN_FIELDS, i)]

    adc_fields = adc_tr_fields + adc_funcs_in_fields

    pricing_fields = [
        i
        for i in fields
        if (i not in adc_fields) and (i not in NOT_ALLOWED_SNAPSHOT_FIELDS)
    ]

    is_parameters_in_adc = any(
        i for i in adc_fields if re.match(ADC_PARAM_IN_FIELDS, i)
    )
    is_parameters_requested = parameters or is_parameters_in_adc

    adc_default_df = _create_default_df(universe, adc_fields)

    # if adc_fields is empty, we need any field to get universe
    _fields = adc_fields or ["TR.RIC"]

    _adc_df = _send_request(
        data_provider=_fundamental_data,
        params={
            "universe": universe,
            "fields": _fields,
            "parameters": parameters,
            "use_field_names_in_headers": use_field_names_in_headers,
        },
        logger=logger,
        default_df=adc_default_df,
    )
    if adc_fields:
        adc_df = _adc_df

    # update universe
    universe = list(_adc_df.get("Instrument", universe))
    _pricing_stream._universe = universe
    _pricing_stream._stream._universe = universe
    _pricing_stream._stream.universe = universe_arg_parser.get_list(universe)

    if not adc_fields:
        _add_flag(adc_df, {"raise_exception": True, "exception": ""})

    if pricing_fields:
        pricing_default_df = _create_default_df(universe, pricing_fields)
        pricing_df = _get_snapshot(
            _pricing_stream, pricing_fields, logger, pricing_default_df
        )
    elif not fields:
        pricing_df = _get_snapshot(_pricing_stream, fields, logger, pricing_df)
    else:
        _add_flag(pricing_df, {"raise_exception": True, "exception": ""})

    _look_for_two_exceptions(pricing_df, adc_df)

    if pricing_df.empty:
        result = adc_df
    elif adc_df.empty:
        result = pricing_df
    elif is_parameters_requested and _both_flags_false(pricing_df, adc_df):
        result = _custom_merge(adc_df, pricing_df)
    else:
        try:
            result = merge(pricing_df, adc_df)
        except pd.errors.MergeError as e:
            logger.debug(
                f"pricing_df=\n"
                f"{pricing_df.to_string()}\n"
                f"adc_df=\n"
                f"{adc_df.to_string()}\n"
            )
            raise e

    columns = get_name_columns_without_datetime(result)
    result[columns] = result[columns].replace([pd.NaT, np.NaN], pd.NA)

    return result


def get_log_string(fields, universe):
    return f"Fields: {fields} for {universe}"


def _send_request(
    data_provider: Callable,
    params: dict,
    logger: "Logger",
    default_df: DataFrame = DataFrame(),
) -> DataFrame:
    fields = params.get("fields", "")
    universe = params["universe"]
    logger.info(f"Requesting {get_log_string(fields, universe)} \n")
    df = default_df

    try:
        response = data_provider(**params).get_data()
    except Exception as e:
        if DEBUG:
            raise e
        logger.exception(f"Failure sending request with {data_provider}")
        df.flags.exception_event = {
            "raise_exception": True,
            "exception": sys.exc_info()[:2],
        }
    else:
        request_messages = response.request_message
        statuses = response.http_status
        if not isinstance(response.request_message, list):
            request_messages = [response.request_message]
            statuses = [response.http_status]

        for request, status in zip(request_messages, statuses):
            path = request.url.path
            cur_universe = path.rsplit("/", 1)[-1]
            if cur_universe not in universe:
                cur_universe = universe
            logger.info(
                f"Request to {path} with {get_log_string(fields, cur_universe)}\n"
                f"status: {status}\n"
            )

        # this check can be removed after all API would return
        # dataframe in response or raise exception
        if response.data.df is not None:
            df = response.data.df
            df.flags.exception_event = {"raise_exception": False, "exception": None}
        else:
            df.flags.exception_event = {
                "raise_exception": True,
                "exception": response.http_status,
            }

    return df


def _rename_column_n_to_column(
    name: str, df: DataFrame, multiindex: bool = False, level: int = 1
) -> None:
    if multiindex:
        occurrences = (
            # fr"^{name}_\d+$" - searching columns like f"{name}_0", f"{name}_1"
            i
            for i in df.columns.levels[level]
            if re.match(rf"^{name}_\d+$", i)
        )
    else:
        occurrences = (i for i in df.columns if re.match(rf"^{name}_\d+$", i))

    df.rename(columns={i: name for i in occurrences}, inplace=True)


def _rename_column_to_column_n(name: str, df: DataFrame) -> None:
    new_names = []
    count = 0
    for i in df.columns:
        if i == name:
            if isinstance(i, tuple):
                column_name = i[1]
                column_name = f"{column_name}_{count}"
                i = (i[0], column_name)
            else:
                i = f"{i}_{count}"
            count += 1
        new_names.append(i)
    df.columns = new_names


def _custom_merge(df_1: DataFrame, df_2: DataFrame) -> DataFrame:
    date_column = "Date"
    instruments_column = "Instrument"

    if "instrument" in df_1:
        df_1.rename(columns={"instrument": instruments_column}, inplace=True)

    duplicated_columns = _find_and_rename_duplicated_columns(df_1)

    if date_column in duplicated_columns:
        date_column = f"{date_column}_0"

    convert_df_columns_to_datetime(df_1, pattern="Date", utc=True, delete_tz=True)

    if instruments_column in df_1 and date_column in df_1:
        grouped = df_1.groupby(instruments_column)
        latest_rows_indexes = grouped[date_column].idxmax(skipna=False)

    else:
        latest_rows_indexes = Series([])

    _remove_empty_rows(latest_rows_indexes)

    latest_info = df_1.loc[latest_rows_indexes]

    if not latest_info.empty:
        mediator = pd.merge(df_2, latest_info, how="outer")
        df_1 = df_1.replace({"": np.NaN})
        result = pd.merge(mediator, df_1, how="right")

    else:
        result = pd.merge(df_2, df_1, how="outer")

    for i in duplicated_columns:
        _rename_column_n_to_column(i, result)

    return result


def _get_snapshot(
    stream: Stream,
    fields: Optional[list],
    logger: "Logger",
    default_df: DataFrame = DataFrame(),
) -> Optional[DataFrame]:
    logger.info(f"Requesting pricing info for fields={fields} via websocket\n")
    df = default_df

    try:
        stream.open(with_updates=False)
        df = stream.get_snapshot(fields=fields)
        stream.close()
    except Exception:
        logger.exception(f"Failure retrieving snapshot for {stream._stream.universe}")
        df.flags.exception_event = {
            "raise_exception": True,
            "exception": sys.exc_info()[:2],
        }
    else:
        # detect if get_snapshot didn't return at least 1 field
        if df.shape[1] == 1:
            msg = f"No data in snapshot for {stream._stream.universe}"
            logger.exception(msg)
            df.flags.exception_event = {"raise_exception": True, "exception": msg}
        else:
            df.flags.exception_event = {"raise_exception": False, "exception": None}

    return df


def _find_and_rename_duplicated_columns(df: DataFrame) -> list:
    counted_columns = Counter(df.columns)
    duplicated_columns = [i for i, n in counted_columns.items() if n > 1]

    for i in duplicated_columns:
        _rename_column_to_column_n(i, df)

    return duplicated_columns


def _remove_empty_rows(series: Series) -> None:
    for name, value in series.items():
        if pd.isna(value):
            series.pop(name)


def _create_default_df(
    universe: list,
    fields: Optional[list] = None,
) -> DataFrame:
    df = DataFrame(
        {
            "Instrument": {n: i for n, i in enumerate(universe)},
            **{i: {} for i in fields},
        }
    )
    return df


def _look_for_two_exceptions(df_1: DataFrame, df_2: DataFrame) -> None:
    if (
        df_1.flags.exception_event["raise_exception"]
        and df_2.flags.exception_event["raise_exception"]
    ):
        ex1 = df_1.flags.exception_event["exception"]
        ex2 = df_2.flags.exception_event["exception"]

        raise RDError(1, f"\nNo data to return, please check errors:\n{ex1}\n{ex2}\n")


def _add_flag(df: DataFrame, flag: Optional[dict] = None) -> None:
    if flag is None:
        flag = {"raise_exception": False, "exception": ""}
    df.flags.exception_event = flag


def _both_flags_false(df_1: DataFrame, df_2: DataFrame) -> bool:
    return (
        not df_1.flags.exception_event["raise_exception"]
        and not df_2.flags.exception_event["raise_exception"]
    )



from ._calendar_adjustment import *
from ._compounding_type import *
from ._market_data_access_denied_fallback import *
from ._zc_interpolation_mode import *
from ._forward_curves_outputs import *
from ._zc_curves_outputs import *
from ._swap_price_side import *
from ..._enums._risk_type import *
from ..._enums._extrapolation_mode import *
from ..._enums._price_side import *
from ..._enums._day_count_basis import *
from ..._enums._interpolation_mode import *
from ..._enums._asset_class import *



# coding: utf8

__all__ = ["Definition"]

from .._base_definition import BaseDefinition
from ._swap_definition import SwapInstrumentDefinition


class Definition(BaseDefinition):
    def __init__(
        self,
        instrument_code=None,
        instrument_tag=None,
        trade_date=None,
        start_date=None,
        end_date=None,
        tenor=None,
        settlement_ccy=None,
        is_non_deliverable=None,
        template=None,
        legs=None,
        fields=None,
        pricing_parameters=None,
        extended_params=None,
    ):
        definition = SwapInstrumentDefinition(
            instrument_tag=instrument_tag,
            trade_date=trade_date,
            start_date=start_date,
            end_date=end_date,
            tenor=tenor,
            settlement_ccy=settlement_ccy,
            is_non_deliverable=is_non_deliverable,
            instrument_code=instrument_code,
            template=template,
            legs=legs,
        )
        super().__init__(
            definition=definition,
            fields=fields,
            pricing_parameters=pricing_parameters,
            extended_params=extended_params,
        )



from enum import Enum

from .._content_type import ContentType
from ..._tools import make_enum_arg_parser, ArgsParser, validate_bool_value


class DataGridType(Enum):
    UDF = "udf"
    RDP = "rdp"


data_grid_types_arg_parser = make_enum_arg_parser(DataGridType)
use_field_names_in_headers_arg_parser = ArgsParser(validate_bool_value)

data_grid_type_value_by_content_type = {
    DataGridType.UDF.value: ContentType.DATA_GRID_UDF,
    DataGridType.RDP.value: ContentType.DATA_GRID_RDP,
}



# coding: utf8

from typing import Optional, Iterable, Union

from ....._tools import validate_types
from ._bond_pricing_parameters import PricingParameters
from ._bond_definition import BondInstrumentDefinition
from ._enums import (
    AdjustInterestToPaymentDate,
    BusinessDayConvention,
    DayCountBasis,
    Direction,
    Frequency,
    IndexCompoundingMethod,
    InterestType,
    DateRollingConvention,
    StubRule,
)
from ._models import AmortizationItem
from .._base_definition import BaseDefinition
from .._contracts_data_provider import bond_instrument_code_arg_parser


class Definition(BaseDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_code : str, optional
        Code to define the bond instrument.
    instrument_tag : str, optional
        User defined string to identify the instrument. It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported.
    end_date : str, optional
        Maturity date of the bond to override. Mandatory if instrument code has not been
        defined and is_perpetual flag has been set to false. In case an instrument code
        has been defined, value comes from bond reference data.
    direction : Direction, optional
        The direction of the leg. Optional for a single leg instrument (like a bond),
        in that case default value is Received. It is mandatory for a multi-instrument
        leg instrument (like Swap or CDS leg).
    interest_type : InterestType, optional
        A flag that indicates whether the leg is fixed or float.
    notional_ccy : str, optional
        The ISO code of the notional currency. Mandatory if instrument code or
        instrument style has not been defined. In case an instrument code/style has been
        defined, value may comes from the reference data.
    notional_amount : float, optional
        The notional amount of the leg at the period start date.
        By default 1,000,000 is used.
    fixed_rate_percent : float, optional
        The fixed coupon rate in percentage. It is mandatory in case of a single leg
        instrument. Otherwise, in case of multi leg instrument, it can be computed as
        the Par rate.
    spread_bp : float, optional
        The spread in basis point that is added to the floating rate index index value.
        By default 0 is used.
    interest_payment_frequency : Frequency or str, optional
        The frequency of the interest payments. Optional if an instrument code/style
        have been defined : in that case, value comes from reference data. Otherwise, it
        is mandatory.
    interest_calculation_method : DayCountBasis or str, optional
        The Day Count Basis method used to calculate the coupon interest payments.
    accrued_calculation_method : DayCountBasis, optional
        The Day Count Basis method used to calculate the accrued interest payments.
        By default, the same value than interest_calculation_method is used.
    payment_business_day_convention : BusinessDayConvention, optional
        The method to adjust dates to a working day.
        In case an instrument code/style has been defined, value comes from bond
        reference data. Otherwise 'ModifiedFollowing' is used.
    payment_roll_convention : DateRollingConvention, optional
        Method to adjust payment dates when they fall at the end of the month (28th of
        February, 30th, 31st). In case an instrument code has been defined,
        value comes from bond reference data. Otherwise, 'SameDay' is used.
    index_reset_frequency : Frequency, optional
        The reset frequency in case the leg Type is Float.
        By default, the IndexTenor is used.
    index_fixing_lag : int, optional
        Defines the number of working days between the fixing date and the start of the
        coupon period ('InAdvance') or the end of the coupon period ('InArrears').
        By default 0 is used.
    first_regular_payment_date : str, optional
        The first regular coupon payment date for leg with an odd first coupon.
    last_regular_payment_date : str, optional
        The last regular coupon payment date for leg with an odd last coupon.
    amortization_schedule : AmortizationItem, optional
        Definition of amortizations.
    payment_business_days : str, optional
        A list of coma-separated calendar codes to adjust dates (e.g. 'EMU' or 'USA').
        By default the calendar associated to notional_ccy is used.
    adjust_interest_to_payment_date : AdjustInterestToPaymentDate, optional
        A flag that indicates if the coupon dates are adjusted to the payment dates.
        By default 'false' is used.
    index_compounding_method : IndexCompoundingMethod, optional
        A flag that defines how the coupon rate is calculated from the reset floating
        rates when the reset frequency is higher than the interest payment frequency
        (e.g. daily index reset with quarterly interest payment).
        By default 'Constant' is used.
    interest_payment_delay : int, optional
        The number of working days between the end of coupon period and the actual
        interest payment date.
        By default no delay (0) is applied.
    stub_rule : StubRule, optional
        The rule that defines whether coupon roll dates are aligned on the  maturity or
        the issue date.
        By default 'Maturity' is used.
    issue_date : str, optional
        Date of issuance of the bond to override. Mandatory if instrument code has not
        been defined. In case an instrument code has been defined, value comes from bond
        reference data.
    first_accrual_date : str, optional
        Date at which bond starts accruing. In case an instrument code has
        been defined, value comes from bond reference data. Otherwise default value is
        the issue date of the bond.
    index_fixing_ric : str, optional
        The RIC that carries the fixing value. This value overrides the RIC associated
        by default with the IndexName and IndexTenor.
    is_perpetual : bool, optional
        Flag the defines wether the bond is perpetual or not in case of user defined
        bond. In case an instrument code has been defined, value comes from
        bond reference data. In case of user defined bond, default value is 'false'.
    template : str, optional
        A reference to a Adfin instrument contract or the Adfin detailed contract.
        Either instrument_code, template, or full definition must be provided.
    fields: list of str, optional
        Contains the list of Analytics that the quantitative analytic service will
        compute.
    pricing_parameters : PricingParameters, optional
        The pricing parameters to apply to this instrument. If pricing
        parameters are not provided at this level parameters defined globally at the
        request level are used. If no pricing parameters are provided globally default
        values apply.
    extended_params : dict, optional
        If necessary other parameters.

    Methods
    -------
    get_data(session=session, on_response=on_response)
        Returns a response to the data platform
    get_data_async(session=None, on_response=None, async_mode=None)
        Returns a response asynchronously to the data platform
    get_stream(session=session, api="")
        Get stream quantitative analytic service subscription

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> definition = rdf.bond.Definition(
     ...    issue_date = "2002-02-28",
     ...    end_date = "2032-02-28",
     ...    notional_ccy = "USD",
     ...    interest_payment_frequency = "Annual",
     ...    fixed_rate_percent = 7,
     ...    interest_calculation_method = rdf.bond.DayCountBasis.DCB_ACTUAL_ACTUAL
     ...)
     >>> response = definition.get_data()

     Using get_data_async
     >>> import asyncio
     >>> task = definition.get_data_async()
     >>> response = asyncio.run(task)

     Using get_stream
     >>> response = definition.get_stream()
    """

    def __init__(
        self,
        instrument_code: Optional[str] = None,
        instrument_tag: Optional[str] = None,
        end_date: Optional[str] = None,
        direction: Optional[Direction] = None,
        interest_type: Optional[InterestType] = None,
        notional_ccy: Optional[str] = None,
        notional_amount: Optional[float] = None,
        fixed_rate_percent: Optional[float] = None,
        spread_bp: Optional[float] = None,
        interest_payment_frequency: Union[Frequency, str] = None,
        interest_calculation_method: Union[DayCountBasis, str] = None,
        accrued_calculation_method: Optional[DayCountBasis] = None,
        payment_business_day_convention: Optional[BusinessDayConvention] = None,
        payment_roll_convention: Optional[DateRollingConvention] = None,
        index_reset_frequency: Optional[Frequency] = None,
        index_fixing_lag: Optional[int] = None,
        first_regular_payment_date: Optional[str] = None,
        last_regular_payment_date: Optional[str] = None,
        amortization_schedule: Optional[AmortizationItem] = None,
        payment_business_days: Optional[str] = None,
        adjust_interest_to_payment_date: Optional[AdjustInterestToPaymentDate] = None,
        index_compounding_method: Optional[IndexCompoundingMethod] = None,
        interest_payment_delay: Optional[int] = None,
        stub_rule: Optional[StubRule] = None,
        issue_date: Optional[str] = None,
        first_accrual_date: Optional[str] = None,
        index_fixing_ric: Optional[str] = None,
        is_perpetual: Optional[bool] = None,
        template: Optional[str] = None,
        fields: Optional[Iterable[str]] = None,
        pricing_parameters: Optional[PricingParameters] = None,
        extended_params: Optional[dict] = None,
    ) -> None:

        if instrument_code:
            instrument_code = bond_instrument_code_arg_parser.get_str(instrument_code)

        validate_types(index_fixing_lag, [int, type(None)], "index_fixing_lag")
        validate_types(
            interest_payment_delay, [int, type(None)], "interest_payment_delay"
        )

        definition = BondInstrumentDefinition(
            accrued_calculation_method=accrued_calculation_method,
            adjust_interest_to_payment_date=adjust_interest_to_payment_date,
            amortization_schedule=amortization_schedule,
            direction=direction,
            index_compounding_method=index_compounding_method,
            index_reset_frequency=index_reset_frequency,
            interest_calculation_method=interest_calculation_method,
            interest_payment_frequency=interest_payment_frequency,
            interest_type=interest_type,
            payment_business_day_convention=payment_business_day_convention,
            payment_roll_convention=payment_roll_convention,
            stub_rule=stub_rule,
            end_date=end_date,
            first_accrual_date=first_accrual_date,
            first_regular_payment_date=first_regular_payment_date,
            fixed_rate_percent=fixed_rate_percent,
            index_fixing_lag=index_fixing_lag,
            index_fixing_ric=index_fixing_ric,
            instrument_code=instrument_code,
            instrument_tag=instrument_tag,
            interest_payment_delay=interest_payment_delay,
            is_perpetual=is_perpetual,
            issue_date=issue_date,
            last_regular_payment_date=last_regular_payment_date,
            notional_amount=notional_amount,
            notional_ccy=notional_ccy,
            payment_business_days=payment_business_days,
            spread_bp=spread_bp,
            template=template,
        )
        super().__init__(
            definition=definition,
            fields=fields,
            pricing_parameters=pricing_parameters,
            extended_params=extended_params,
        )



# coding: utf8
# contract_gen 2020-06-03 11:34:39.450949

__all__ = ["Status"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class Status(Enum):
    NOT_APPLICABLE = "NotApplicable"
    USER = "User"
    DATA = "Data"
    COMPUTED = "Computed"
    ERROR = "Error"
    NONE = "None"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(Status, _STATUS_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_STATUS_VALUES_IN_LOWER_BY_STATUS, some)


_STATUS_VALUES = tuple(t.value for t in Status)
_STATUS_VALUES_IN_LOWER_BY_STATUS = {
    name.lower(): item for name, item in Status.__members__.items()
}



# coding: utf8

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class PriceSide(Enum):
    """
    Quoted price side to use for pricing Analysis:
        Bid(Bid value),
        Ask(Ask value),
        Mid(Mid value)
    """

    ASK = "Ask"
    BID = "Bid"
    LAST = "Last"
    MID = "Mid"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(PriceSide, _PRICE_SIDE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_PRICE_SIDE_VALUES_IN_LOWER_BY_PRICE_SIDE, some)


_PRICE_SIDE_VALUES = tuple(t.value for t in PriceSide)
_PRICE_SIDE_VALUES_IN_LOWER_BY_PRICE_SIDE = {
    name.lower(): item for name, item in PriceSide.__members__.items()
}



import threading
from concurrent.futures import wait, ThreadPoolExecutor
from typing import TYPE_CHECKING, Dict, List

from .event import StreamCxnEvent
from .null_stream_connection import NullStreamConnection
from ...content._content_type import ContentType
from ..._core.session.event_code import EventCode

if TYPE_CHECKING:
    from ._stream_factory import StreamDetails
    from .._stream import StreamConnection
    from ..._core.session import Session

StreamConnections = List["StreamConnection"]


def on_event_connecting(cxn: "StreamConnection", message: dict = None):
    on_event_listener(cxn, EventCode.StreamConnecting, message)


def on_event_connected(cxn: "StreamConnection", message: dict = None):
    on_event_listener(cxn, EventCode.StreamConnected, message)


def on_event_disconnected(cxn: "StreamConnection", message: dict = None):
    on_event_listener(cxn, EventCode.StreamDisconnected, message)


def on_event_listener(cxn: "StreamConnection", event: EventCode, message: dict):
    cxn.session._call_on_event(event, message)


def add_listeners(cxn: "StreamConnection") -> None:
    cxn.on(StreamCxnEvent.CONNECTING, on_event_connecting)
    cxn.on(StreamCxnEvent.CONNECTED, on_event_connected)
    cxn.on(StreamCxnEvent.DISCONNECTED, on_event_disconnected)


def remove_listeners(cxn: "StreamConnection") -> None:
    cxn.remove_listener(StreamCxnEvent.CONNECTING, on_event_connecting)
    cxn.remove_listener(StreamCxnEvent.CONNECTED, on_event_connected)
    cxn.remove_listener(StreamCxnEvent.DISCONNECTED, on_event_disconnected)


class CacheItem:
    def __init__(
        self, cxn: "StreamConnection", details: "StreamDetails", owner: dict
    ) -> None:
        self.api_type_as_str = details.api_type_as_str
        self.owner = owner
        self.cxn = cxn
        self.number_in_use = 0

        add_listeners(self.cxn)

    @property
    def is_using(self):
        return self.number_in_use > 0

    def inc_use(self):
        self.number_in_use += 1

    def dec_use(self):
        if self.number_in_use == 0:
            raise ValueError("CacheItem: number_in_use cannot be less 0")

        self.number_in_use -= 1

        if self.number_in_use == 0 and self.cxn.is_disconnecting:
            self.dispose()

    def dispose(self):
        self.number_in_use = -1
        self.owner.pop(self.api_type_as_str)
        self.api_type_as_str = None
        self.owner = None

        remove_listeners(self.cxn)
        cxn = self.cxn
        cxn.dispose()
        cxn.join()
        self.cxn = None

    def __str__(self) -> str:
        if self.cxn:
            name = self.cxn.name
        else:
            name = "disposed"
        return f"CacheItem(cxn={name}, number_in_use={self.number_in_use})"


class StreamCxnCache(object):
    def __init__(self) -> None:
        self._cache: Dict["Session", Dict[str, CacheItem]] = {}
        self._lock = threading.Lock()

    def has_cxn(self, session: "Session", details: "StreamDetails") -> bool:
        item = self._cache.get(session, {}).get(details.api_type_as_str)
        return bool(item)

    def get_cxn(
        self, session: "Session", details: "StreamDetails"
    ) -> "StreamConnection":

        with self._lock:

            content_type = details.content_type
            protocol_type = details.protocol_type

            if not self.has_cxn(session, details):
                from ._stream_factory import create_stream_cxn

                cxn = create_stream_cxn(details, session)
                cxn.start()
                self._add_cxn(cxn, session, details)
                session.debug(
                    f" + StreamCxnCache created new connection "
                    f"(id={cxn.id}, daemon={cxn.daemon}, content_type={content_type}, "
                    f"protocol_type={protocol_type})"
                )

            item = self._get_cache_item(session, details)
            cxn = item.cxn

            session.debug(
                f"StreamCxnCache wait for connection "
                f"(id={cxn.id}, content_type={content_type}, "
                f"protocol_type={protocol_type})"
            )

            wait([cxn.prepared])

            if cxn.is_disposed:
                raise AssertionError(f"Cannot prepare connection {cxn}")

            elif cxn.prepared.result() is True:
                item.inc_use()
                session.debug(f" <=== StreamCxnCache connection id={cxn.id} is ready")

            elif cxn.prepared.result() is False:
                session.error(
                    f"StreamCxnCache cannot get cxn"
                    f"(id={cxn.id}, content_type={content_type}, "
                    f"protocol_type={protocol_type})"
                    f", because it is disposed."
                )
                if self.has_cxn(session, details):
                    self.del_cxn(cxn, session, details)

                cxn.join()
                cxn = NullStreamConnection()

            else:
                raise AssertionError(f"Don't know what to do")

            return cxn

    def release(self, session: "Session", details: "StreamDetails") -> None:
        content_type = details.content_type
        if not self.has_cxn(session, details):
            raise ValueError(
                f"Cannot release stream connection, "
                f"because it’s not in the cache "
                f"(content_type={content_type}, session={session})"
            )

        item_by_api_type = self._cache[session]
        item = item_by_api_type[details.api_type_as_str]
        item.dec_use()
        session.debug(
            f" ===> StreamCxnCache release (item={item},\n"
            f"\t\tcontent_type={content_type},\n"
            f"\t\tsession={session})"
        )

    def del_cxn(
        self, cxn: "StreamConnection", session: "Session", details: "StreamDetails"
    ) -> None:
        content_type = details.content_type

        if not cxn:
            raise ValueError(
                f"Cannot delete stream connection, "
                f"because it is empty (content_type={content_type}, "
                f"cxn={cxn}, session={session})"
            )

        if not self.has_cxn(session, details):
            raise ValueError(
                f"Cannot delete stream connection, "
                f"because already deleted (content_type={content_type}, "
                f"cxn={cxn}, session={session})"
            )

        item_by_api_type = self._cache[session]
        item = item_by_api_type[details.api_type_as_str]
        if item.is_using:
            raise AssertionError(
                f"Cannot delete stream connection, "
                f"because it is using (content_type={content_type}, "
                f"cxn={cxn}, session={session})"
            )

        cached_cxn = item.cxn

        if cxn is not cached_cxn:
            raise ValueError(
                f"Cannot delete stream connection, "
                f"because cxn is not the same \n"
                f"(cxn={cxn} != cached_cxn={cached_cxn},"
                f"content_type={content_type}, session={session})"
            )

        item.dispose()

    def has_cxns(self, session: "Session") -> bool:
        item_by_content_type = self._cache.get(session, {})
        has_cxns = bool(item_by_content_type.values())
        return has_cxns

    def get_cxns(self, session: "Session") -> StreamConnections:
        item_by_content_type = self._cache.get(session, {})
        return [item.cxn for item in item_by_content_type.values()]

    def close_cxns(self, session: "Session") -> None:
        def _close_cxn(item):
            if item.is_using:
                item.cxn.disconnect()

            else:
                item.cxn.dispose()
                item.dispose()

        with ThreadPoolExecutor(thread_name_prefix="CloseCxns-Thread") as pool:
            pool.map(_close_cxn, self._get_cache_items(session))

        self._cache.pop(session, None)

    def _add_cxn(
        self, cxn: "StreamConnection", session: "Session", details: "StreamDetails"
    ) -> CacheItem:
        content_type = details.content_type

        if not cxn:
            raise ValueError(
                f"Cannot add stream connection, "
                f"because it is empty: content_type={content_type}, "
                f"cxn={cxn}, session={session}"
            )

        if self.has_cxn(session, details):
            raise ValueError(
                f"Cannot add stream connection, "
                f"because already added: content_type={content_type}, "
                f"cxn={cxn}, session={session}"
            )

        owner = self._cache.setdefault(session, {})
        item = CacheItem(cxn, details, owner)
        owner[details.api_type_as_str] = item
        return item

    def _get_cache_items(self, session: "Session") -> List[CacheItem]:
        item_by_content_type = self._cache.get(session, {})
        return [item for item in item_by_content_type.values()]

    def _get_cache_item(
        self, session: "Session", details: "StreamDetails"
    ) -> CacheItem:
        cache_item = self._cache[session][details.api_type_as_str]
        return cache_item

    def is_cxn_alive(self, session: "Session", content_type: ContentType) -> bool:
        from ._stream_factory import content_type_to_details

        details = content_type_to_details(content_type)
        is_alive = False
        if self.has_cxn(session, details):
            item = self._get_cache_item(session, details)
            is_alive = item.cxn.is_alive()

        return is_alive


stream_cxn_cache: StreamCxnCache = StreamCxnCache()



# coding: utf8

from enum import Enum, unique
from ...._enums._common_tools import (
    _convert_to_str,
    _normalize,
)


@unique
class UnderlyingType(Enum):
    """
    Underlying type of the option.
    Possible values:
        - Eti
        - Fx
    """

    ETI = "Eti"
    FX = "Fx"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(UnderlyingType, _UNDERLYING_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_UNDERLYING_TYPE_VALUES_IN_LOWER_BY_UNDERLYING_TYPE, some)


_UNDERLYING_TYPE_VALUES = tuple(t.value for t in UnderlyingType)
_UNDERLYING_TYPE_VALUES_IN_LOWER_BY_UNDERLYING_TYPE = {
    name.lower(): item for name, item in UnderlyingType.__members__.items()
}



from typing import Optional, TYPE_CHECKING

from ._stream_facade import Stream
from .._ipa_content_provider import IPAContentProviderLayer
from ..._content_type import ContentType

if TYPE_CHECKING:
    from ...._core.session import Session


class BaseDefinition(IPAContentProviderLayer):
    def __init__(self, **kwargs) -> None:
        super().__init__(ContentType.CONTRACTS, **kwargs)

    def get_stream(
        self,
        session: Optional["Session"] = None,
    ) -> Stream:
        """
        Returns a streaming quantitative analytic service subscription

        Parameters
        ----------
        session : Session, optional
            Means the default session will be used

        Returns
        -------
        Stream

        Raises
        ------
        AttributeError
            If user didn't set default session.
        """
        definition = self._kwargs.get("definition")
        instrument_type = definition.get_instrument_type()
        json = definition.get_json()

        pricing_parameters = self._kwargs.get("pricing_parameters")

        definition = {
            "instrumentType": instrument_type,
            "instrumentDefinition": json,
        }

        if pricing_parameters:
            definition["pricingParameters"] = pricing_parameters.get_json()

        stream = Stream(
            session=session,
            fields=self._kwargs.get("fields"),
            universe=definition,
            extended_params=self._kwargs.get("extended_params"),
        )
        return stream

    def __eq__(self, other):
        definition = self._kwargs.get("definition")
        return definition == other

    def __repr__(self):
        repr_str = super().__repr__()
        new_str = f" {{name='{self._kwargs.get('definition')}'}}>"
        repr_str = repr_str.replace(">", new_str)
        return repr_str


class FCBaseDefinition(IPAContentProviderLayer):
    def __init__(self, **kwargs) -> None:
        super().__init__(ContentType.CONTRACTS, **kwargs)

    def __eq__(self, other):
        definition = self._kwargs.get("definition")
        return definition == other

    def __repr__(self):
        repr_str = super().__repr__()
        new_str = f" {{name='{self._kwargs.get('definition')}'}}>"
        repr_str = repr_str.replace(">", new_str)
        return repr_str



import os

import requests

from ._tools import path_join, create_dirs_if_no_exists, get_correct_filename
from ..._core.log_reporter import _LogReporter


class _FileDownloader(_LogReporter):
    def __init__(self, logger):
        super().__init__(logger=logger)

    def download(self, url, filename_ext, path="") -> str:
        if path and not os.path.exists(path):
            self._debug(f"creating dir '{path}'")
            create_dirs_if_no_exists(path)

        self._debug(f"start downloading file by url {url}")

        r = requests.get(url, allow_redirects=True, stream=True)
        if not r.ok:
            self._error(f"url is not valid, status_code: {r.status_code}")
            raise ValueError(f"url is not valid, status_code: {r.status_code}")

        filepath = path_join(path, get_correct_filename(filename_ext))
        self._debug(f"save file to {filepath}")
        with open(filepath, "wb") as f:
            for data in r.iter_content():
                f.write(data)

        r.close()

        self._debug("file downloaded successfully")

        return filepath



import socket


class DacsParams(object):
    """DacsParams object."""

    def __init__(self, *args, **kwargs):
        self.deployed_platform_username = kwargs.get(
            "deployed_platform_username", "user"
        )
        self.dacs_application_id = kwargs.get("dacs_application_id", "256")
        self.dacs_position = kwargs.get("dacs_position")
        if self.dacs_position in [None, ""]:
            try:
                position_host = socket.gethostname()
                self.dacs_position = "{}/{}".format(
                    socket.gethostbyname(position_host), position_host
                )
            except socket.gaierror:
                self.dacs_position = "127.0.0.1/net"
        self.authentication_token = kwargs.get("authentication_token")



__all__ = (
    "Adjustments",
    "events",
    "EventTypes",
    "Intervals",
    "MarketSession",
    "summaries",
)

from . import events, summaries
from ._hp_data_provider import EventTypes, Adjustments, MarketSession
from .._intervals import Intervals



from itertools import zip_longest
from typing import Any

import numpy
import pandas as pd
from pandas import DataFrame

from ._outputs import Outputs
from ..._content_provider import ErrorParser
from ...._tools import (
    ArgsParser,
    iterable,
    make_enum_arg_parser,
    fields_arg_parser,
)
from ...._tools._dataframe import convert_column_df_to_datetime
from ....delivery._data._data_provider import ContentValidator, ValidatorContainer
from ....delivery._data._data_provider import (
    RequestFactory,
    DataProvider,
    ResponseFactory,
    Data as ContentData,
)
from ....delivery.endpoint_request import RequestMethod


# --------------------------------------------------------------------------------------
#   Content data validator
# --------------------------------------------------------------------------------------


class ContractsContentValidator(ContentValidator):
    def validate(self, data, *args, **kwargs):
        content_data = data.get("content_data")
        status = content_data.get("status", "")

        if content_data is None:
            data["error_code"] = 1
            data["error_message"] = "Content data is None"
            return False

        elif status == "Error":
            data["error_code"] = content_data.get("code")
            data["error_message"] = content_data.get("message")
            return False

        else:
            return self._is_valid_if_have_all_errors(data, content_data)

    def _is_valid_if_have_all_errors(self, data: dict, content_data: dict) -> bool:
        headers = content_data.get("headers", [])
        datas = content_data.get("data", [])
        err_codes = []
        err_msgs = []

        for header, *data_items in zip(headers, *datas):
            header_name = header["name"]

            if "ErrorCode" == header_name:
                err_codes = data_items

            elif "ErrorMessage" == header_name:
                err_msgs = data_items

        counter = len(datas) or 1  # because datas can be empty list
        if err_codes or err_msgs:
            for err_code, err_msg in zip_longest(err_codes, err_msgs, fillvalue=None):
                if err_code or err_msg:
                    counter -= 1
                    error_codes = data.setdefault("error_code", [])
                    error_codes.append(err_code)
                    error_messages = data.setdefault("error_message", [])
                    error_messages.append(err_msg)

        if counter == 0:
            return False

        return True


# ---------------------------------------------------------------------------
#   Content data
# ---------------------------------------------------------------------------


def convert_data_items_to_datetime(df: pd.DataFrame, headers) -> pd.DataFrame:
    for index, header in enumerate(headers):
        header_type = header.get("type", "")
        if header_type == "DateTime" or header_type == "Date":
            convert_column_df_to_datetime(df, index)

    return df


def financial_contracts_build_df(raw: dict, **kwargs) -> pd.DataFrame:
    """
    Convert "data" from raw response bond to dataframe format
    """
    data = raw.get("data")
    headers = raw.get("headers")
    if data:
        numpy_array = numpy.array(data, dtype=object)
        df = DataFrame(numpy_array)
        df = convert_data_items_to_datetime(df, headers)
        df.columns = [header["name"] for header in headers]
        if not df.empty:
            df.fillna(pd.NA, inplace=True)
            df = df.convert_dtypes()
    else:
        df = DataFrame([], columns=[])
    return df


class Data(ContentData):
    """
    This class is designed for storing and managing the response instrument data
    """

    def __init__(self, raw, **kwargs):
        super().__init__(raw, **kwargs)
        self._analytics_headers = None
        self._analytics_data = None
        self._analytics_market_data = None
        self._analytics_statuses = None
        if raw:
            #   get headers
            self._analytics_headers = raw.get("headers")
            #   get data
            self._analytics_data = raw.get("data")
            #   get marketData
            self._analytics_market_data = raw.get("marketData")
            #   get statuses
            self._analytics_statuses = raw.get("statuses")

    @property
    def analytics_headers(self):
        return self._analytics_headers

    @property
    def analytics_data(self):
        return self._analytics_data

    @property
    def analytics_market_data(self):
        return self._analytics_market_data

    @property
    def analytics_statuses(self):
        return self._analytics_statuses

    @property
    def marketdata_df(self):
        """
        Convert "marketData" from raw response bond to dataframe format
        """
        return None


# ---------------------------------------------------------------------------
#   Request factory
# ---------------------------------------------------------------------------


class ContractsRequestFactory(RequestFactory):
    def get_request_method(self, *_, **kwargs):
        return kwargs.get("method", RequestMethod.POST)

    def get_body_parameters(self, *_, **kwargs):
        input_universe = kwargs.get("universe")
        input_universe = input_universe or kwargs.get("definition") or []

        if not isinstance(input_universe, list):
            input_universe = [input_universe]

        universe = []
        # convert universe's objects into json
        for i, item in enumerate(input_universe):
            # item can be a tuple (instrument, pricing_parameters)
            if iterable(item):
                data = get_data(*item)

            else:
                data = get_data(item)

            universe.append(data)

        body_parameters = {"universe": universe}

        fields = kwargs.get("fields") or []
        if fields:
            fields = fields_arg_parser.get_list(fields)
            body_parameters["fields"] = fields

        outputs = kwargs.get("outputs")
        if outputs:
            outputs = outputs_contracts_arg_parser.get_list(outputs)
            body_parameters["outputs"] = outputs

        pricing_parameters = kwargs.get("pricing_parameters")
        if pricing_parameters:
            body_parameters["pricingParameters"] = pricing_parameters.get_json()

        return body_parameters


def get_data(definition, pricing_parameters=None):
    fields = None
    if hasattr(definition, "_kwargs"):
        kwargs = getattr(definition, "_kwargs")
        definition = kwargs.get("definition")
        fields = kwargs.get("fields")
        pricing_parameters = kwargs.get("pricing_parameters")

    data = {
        "instrumentType": definition.get_instrument_type(),
        "instrumentDefinition": definition.get_json(),
    }

    if fields:
        fields = fields_arg_parser.get_list(fields)
        data["fields"] = fields

    if pricing_parameters:
        data["pricingParameters"] = pricing_parameters.get_json()

    return data


def process_universe_item(item, definition_class):
    # item can be empty string ""
    if item and isinstance(item, str):
        item = definition_class(item)

    elif isinstance(item, definition_class):
        return item

    else:
        raise ValueError(
            f"Invalid type of universe, expected str: {type(item)} is given"
        )

    return item


def process_universe(universe, definition_class, pricing_parameters_class=None):
    if iterable(universe):
        for i, item in enumerate(universe):

            if iterable(item):
                item, pricing_parameters = item

                if pricing_parameters_class and not isinstance(
                    pricing_parameters, pricing_parameters_class
                ):
                    raise ValueError(
                        f"Invalid type of pricing parameters, "
                        f"expected {pricing_parameters_class}: "
                        f"{type(pricing_parameters)} is given"
                    )

                item = process_universe_item(item, definition_class)
                item = (item, pricing_parameters)

            else:
                item = process_universe_item(item, definition_class)

            universe[i] = item

    else:
        universe = process_universe_item(universe, definition_class)

    return universe


def process_bond_instrument_code(code: Any) -> str:
    if code is None or isinstance(code, str):
        return code
    else:
        raise ValueError(
            f"Invalid type of instrument_code, string is expected."
            f"type: {type(code)} is given"
        )


universe_contracts_arg_parser = ArgsParser(process_universe)
outputs_contracts_arg_parser = make_enum_arg_parser(Outputs)
bond_instrument_code_arg_parser = ArgsParser(process_bond_instrument_code)

# ---------------------------------------------------------------------------
#   Data provider
# ---------------------------------------------------------------------------

contracts_data_provider = DataProvider(
    request=ContractsRequestFactory(),
    response=ResponseFactory(data_class=Data),
    validator=ValidatorContainer(content_validator=ContractsContentValidator()),
    parser=ErrorParser(),
)



from ..._curves._enums import (
    DayCountBasis,
    CalendarAdjustment,
    ZcInterpolationMode,
    PriceSide,
    MarketDataAccessDeniedFallback,
    CompoundingType,
    ExtrapolationMode,
    RiskType,
    AssetClass,
)



# coding: utf8

__all__ = ["CalibrationStrategy"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class CalibrationStrategy(Enum):
    ALL_MATURITY = "AllMaturity"
    DEFAULT = "Default"
    MGB_CALIBRATION = "MGBCalibration"
    MATURITY_BY_MATURITY = "MaturityByMaturity"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(CalibrationStrategy, _CALIBRATION_STRATEGY_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _CALIBRATION_STRATEGY_VALUES_IN_LOWER_BY_CALIBRATION_STRATEGY, some
        )


_CALIBRATION_STRATEGY_VALUES = tuple(t.value for t in CalibrationStrategy)
_CALIBRATION_STRATEGY_VALUES_IN_LOWER_BY_CALIBRATION_STRATEGY = {
    name.lower(): item for name, item in CalibrationStrategy.__members__.items()
}



__all__ = (
    "Response",
    "surface_cap",
    "surface_eti",
    "surface_fx",
    "surface_swaption",
    "surfaces",
)

from . import surface_cap
from . import surface_eti
from . import surface_fx
from . import surface_swaption
from . import surfaces
from ....delivery._data._data_provider import Response



# coding: utf8

__all__ = ["DividendExtrapolation"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class DividendExtrapolation(Enum):
    CST_EXTRAPOL = "Cst_Extrapol"
    LINEAR_EXTRAPOL = "Linear_Extrapol"
    POWER_GROWTH_EXTRAPOL = "PowerGrowth_Extrapol"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            DividendExtrapolation, _DIVIDEND_EXTRAPOLATION_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _DIVIDEND_EXTRAPOLATION_VALUES_IN_LOWER_BY_DIVIDEND_EXTRAPOLATION, some
        )


_DIVIDEND_EXTRAPOLATION_VALUES = tuple(t.value for t in DividendExtrapolation)
_DIVIDEND_EXTRAPOLATION_VALUES_IN_LOWER_BY_DIVIDEND_EXTRAPOLATION = {
    name.lower(): item for name, item in DividendExtrapolation.__members__.items()
}



from typing import Union

import numpy as np
from scipy import interpolate

from . import CurvePoint
from ..._content_provider import value_arg_parser
from ..._enums import Axis


class Curve:
    """
    Methods
    -------
        get_point(x, y): CurvePoint

    Examples
    -------
    >>> from refinitiv.data.content.ipa.curves import forward_curve
    >>> definition = forward_curve.Definition(
    ...    curve_definition=forward_curve.SwapZcCurveDefinition(
    ...        currency="EUR",
    ...        index_name="EURIBOR",
    ...        name="EUR EURIBOR Swap ZC Curve",
    ...        discounting_tenor="OIS",
    ...
    ...    ),
    ...    forward_curve_definitions=[
    ...        forward_curve.ForwardCurveDefinition(
    ...            index_tenor="3M",
    ...            forward_curve_tag="ForwardTag",
    ...            forward_start_date="2021-02-01",
    ...            forward_curve_tenors=[
    ...                "0D",
    ...                "1D",
    ...                "2D",
    ...                "3M",
    ...                "6M",
    ...                "9M",
    ...                "1Y",
    ...                "2Y",
    ...                "3Y",
    ...                "4Y",
    ...                "5Y",
    ...                "6Y",
    ...                "7Y",
    ...                "8Y",
    ...                "9Y",
    ...                "10Y",
    ...                "15Y",
    ...                "20Y",
    ...                "25Y"
    ...            ]
    ...        )
    ...    ]
    ...)
    >>> response = definition.get_data()
    >>> curve = response.data.curve
    >>> point = curve.get_point("2021-02-02", Axis.X)
    """

    def __init__(self, x: np.array, y: np.array) -> None:
        self._x = x
        self._y = y

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    def get_point(
        self,
        value: Union[str, float, np.datetime64],
        axis: Axis,
    ) -> CurvePoint:
        """
        Parameters
        ----------
        value: str, float, np.datetime64

        axis: Axis

        Returns
        -------
        CurvePoint

        Raises
        -------
        ValueError
            If cannot identify axis
        """
        value = value_arg_parser.parse(value)

        is_value_on_axis_x = axis is Axis.X
        is_value_on_axis_y = axis is Axis.Y

        value_as_float = value_arg_parser.get_float(value)

        if is_value_on_axis_x:
            f = interpolate.interp1d(
                x=self._x.astype(float),
                y=self._y.astype(float),
            )
            x = value
            y = f(x=value_as_float)
            y = float(y)

        elif is_value_on_axis_y:
            f = interpolate.interp1d(
                x=self._y.astype(float),
                y=self._x.astype(float),
            )
            x = f(x=value_as_float)
            x = float(x)
            y = value

        else:
            raise ValueError(f"Cannot identify axis={axis}")

        return CurvePoint(x, y)

    def __str__(self) -> str:
        x = self._x
        if np.iterable(self._x):
            x = ", ".join(str(i) for i in self._x)

        y = self._y
        if np.iterable(self._y):
            y = ", ".join(str(i) for i in self._y)

        return f"X={x}\nY={y}"


class ForwardCurve(Curve):
    def __init__(self, x: np.array, y: np.array, **kwargs) -> None:
        super().__init__(x, y)
        self.forward_curve_tag = kwargs.get("forwardCurveTag")
        self.forward_start = kwargs.get("forwardStart")
        self.index_tenor = kwargs.get("indexTenor")


class ZcCurve(Curve):
    def __init__(self, x: np.array, y: np.array, index_tenor, **kwargs) -> None:
        super().__init__(x, y)
        self.index_tenor = index_tenor
        self.is_discount_curve = kwargs.get("isDiscountCurve")



# coding: utf-8

from typing import Callable, Optional, TYPE_CHECKING, Any

from ._protocol_type import ProtocolType
from ._stream_listener import OMMStreamListener, StreamEvent
from .stream import (
    Stream,
    update_message_with_extended_params,
)
from ...content._types import OptStr, ExtendedParams, Strings
from ..._tools import cached_property

if TYPE_CHECKING:
    from ._stream_factory import StreamDetails
    from . import StreamConnection
    from ..._core.session import Session


class _OMMStream(Stream, OMMStreamListener["_OMMStream"]):
    _message = None
    _code = None
    _with_updates: bool = True

    def __init__(
        self,
        stream_id: int,
        session: "Session",
        name: str,
        details: "StreamDetails",
        domain: OptStr = "MarketPrice",
        service: OptStr = None,
        fields: Optional[Strings] = None,
        key: Optional[dict] = None,
        extended_params: ExtendedParams = None,
        on_refresh: Optional[Callable] = None,
        on_status: Optional[Callable] = None,
        on_update: Optional[Callable] = None,
        on_complete: Optional[Callable] = None,
        on_error: Optional[Callable] = None,
    ) -> None:
        Stream.__init__(self, stream_id, session, details)
        OMMStreamListener.__init__(
            self,
            logger=session.logger(),
            on_refresh=on_refresh,
            on_status=on_status,
            on_update=on_update,
            on_complete=on_complete,
            on_error=on_error,
        )
        self._name = name
        self._service = service
        self._fields = fields or []
        self._domain = domain
        self._key = key
        self._extended_params = extended_params

    @property
    def name(self) -> str:
        return self._name

    @property
    def protocol_type(self) -> ProtocolType:
        return ProtocolType.OMM

    @property
    def code(self):
        return self._code

    @property
    def message(self):
        return self._message

    @cached_property
    def open_message(self):
        open_message = {
            "ID": self.id,
            "Domain": self._domain,
            "Streaming": self._with_updates,
            "Key": {},
        }

        if self._key:
            open_message["Key"] = self._key

        open_message["Key"]["Name"] = self.name

        if self._service:
            open_message["Key"]["Service"] = self._service

        if self._fields:
            open_message["View"] = self._fields

        if self._extended_params:
            open_message = update_message_with_extended_params(
                open_message, self._extended_params
            )

        return open_message

    @cached_property
    def close_message(self):
        return {"ID": self.id, "Type": "Close"}

    def _do_open(self, *args, **kwargs):
        self._with_updates = kwargs.get("with_updates", True)

        self._initialize_cxn()

        self._cxn.on(self._event.refresh_by_id, self._on_stream_refresh)
        self._cxn.on(self._event.update_by_id, self._on_stream_update)
        self._cxn.on(self._event.status_by_id, self._on_stream_status)
        self._cxn.on(self._event.complete_by_id, self._on_stream_complete)
        self._cxn.on(self._event.error_by_id, self._on_stream_error)

        super()._do_open(*args, **kwargs)

    def _dispose(self):
        self._debug(f"{self._classname} disposing [d]")

        self._cxn.remove_listener(self._event.refresh_by_id, self._on_stream_refresh)
        self._cxn.remove_listener(self._event.update_by_id, self._on_stream_update)
        self._cxn.remove_listener(self._event.status_by_id, self._on_stream_status)
        self._cxn.remove_listener(self._event.complete_by_id, self._on_stream_complete)
        self._cxn.remove_listener(self._event.error_by_id, self._on_stream_error)

        self._release_cxn()
        self._debug(f"{self._classname} disposed [D]")

    def _do_on_stream_refresh(self, cxn: "StreamConnection", *args) -> Any:
        message = args[0]
        message_state = message.get("State", {})
        self._code = message_state.get("Stream", "")
        self._message = message_state.get("Text", "")
        return message

    def _on_stream_status(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.STATUS, originator, *args)

        if self.is_open:
            self.dispatch_complete(originator, *args)

    def _do_on_stream_status(self, cxn: "StreamConnection", *args) -> Any:
        message = args[0]
        message_state = message.get("State", {})
        stream_state = message_state.get("Stream", "")
        code_state = message_state.get("Code", "")

        self._code = stream_state
        self._message = message_state.get("Text", "")

        if stream_state == "Closed":
            self._debug(
                f"{self._classname} received a closing message, "
                f"message_state={message_state}, state={self.state}"
            )

            if self.is_opening:
                self.dispatch_error(self, message)

            elif self.is_opened and code_state in {"AlreadyOpen", "NotEntitled"}:
                self.dispatch_error(self, message)

            elif self.is_closing:
                self._closed.set_result(True)

            else:
                raise ValueError(
                    f"{self._classname} Don't know what to do, state={self.state}, message={message}"
                )

        return message

    def _on_stream_complete(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.COMPLETE, originator, *args)

        if self.is_opening:
            not self._opened.done() and self._opened.set_result(True)



# coding: utf8
# contract_gen 2020-05-18 08:30:59.202852

__all__ = ["IndexCompoundingMethod"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class IndexCompoundingMethod(Enum):
    """
    A flag that defines how the coupon rate is calculated from the reset floating
    rates when the reset frequency is higher than the interest payment frequency
    (e.g. daily index reset with quarterly interest payment).

    The possible values are:
        - Compounded (uses the compounded average rate from multiple fixings),
        - Average (uses the arithmetic average rate from multiple fixings),
        - Constant (uses the last published rate among multiple fixings),
        - AdjustedCompounded (uses Chinese 7-day repo fixing),
        - MexicanCompounded (uses Mexican Bremse fixing).
    """

    ADJUSTED_COMPOUNDED = "AdjustedCompounded"
    AVERAGE = "Average"
    COMPOUNDED = "Compounded"
    CONSTANT = "Constant"
    MEXICAN_COMPOUNDED = "MexicanCompounded"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            IndexCompoundingMethod, _INDEX_COMPOUNDING_METHOD_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _INDEX_COMPOUNDING_METHOD_VALUES_IN_LOWER_BY_INDEX_COMPOUNDING_METHOD, some
        )


_INDEX_COMPOUNDING_METHOD_VALUES = tuple(t.value for t in IndexCompoundingMethod)
_INDEX_COMPOUNDING_METHOD_VALUES_IN_LOWER_BY_INDEX_COMPOUNDING_METHOD = {
    name.lower(): item for name, item in IndexCompoundingMethod.__members__.items()
}



# coding: utf8

from enum import Enum, unique

from ..._enums._common_tools import _convert_to_str, _normalize


@unique
class MainConstituentAssetClass(Enum):
    FUTURES = "Futures"
    SWAP = "Swap"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            MainConstituentAssetClass, _MAIN_CONSTITUENT_ASSET_CLASS_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _MAIN_CONSTITUENT_ASSET_CLASS_VALUES_IN_LOWER_BY_MAIN_CONSTITUENT_ASSET_CLASS,
            some,
        )


_MAIN_CONSTITUENT_ASSET_CLASS_VALUES = tuple(t.value for t in MainConstituentAssetClass)
_MAIN_CONSTITUENT_ASSET_CLASS_VALUES_IN_LOWER_BY_MAIN_CONSTITUENT_ASSET_CLASS = {
    name.lower(): item for name, item in MainConstituentAssetClass.__members__.items()
}



# coding: utf8

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class AdjustInterestToPaymentDate(Enum):
    ADJUSTED = "Adjusted"
    UNADJUSTED = "Unadjusted"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            AdjustInterestToPaymentDate, _ADJUST_INTEREST_TO_PAYMENT_DATE_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _ADJUST_INTEREST_TO_PAYMENT_DATE_VALUES_IN_LOWER_BY_ADJUST_INTEREST_TO_PAYMENT_DATE,
            some,
        )


_ADJUST_INTEREST_TO_PAYMENT_DATE_VALUES = tuple(
    t.value for t in AdjustInterestToPaymentDate
)
_ADJUST_INTEREST_TO_PAYMENT_DATE_VALUES_IN_LOWER_BY_ADJUST_INTEREST_TO_PAYMENT_DATE = {
    name.lower(): item for name, item in AdjustInterestToPaymentDate.__members__.items()
}
